/* Individual-based model of Galton-Watson process */
/* F. Takasu 2008 March 9 */
/* An individual gives N birth(s) 
	N = 0 with prob. P0
	N = 1 with prob. P1
	N = 2 with prob. P2
	N = 3 with prob. P3 (P0 + P1 + P2 + P3 = 1),
	
	and dies with prob. 1 - SURVIVAL */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#define p0 0.25		/* Prob. 0 offspring */
#define p1 0.25		/* Prob. 1 offspring */
#define p2 0.25		/* Prob. 2 offspring */
#define p3 0.25		/* Prob. 3 offspring */

#define SURVIVAL	0.3		/* Prob. of survive to next year */

#define TRIAL	100	/* Number of realizations */
#define STEP	10	/* Time steps */

#define INIT_POP_SIZE	1	/* Initial pop. size */


typedef struct individual INDIVIDUAL;
struct individual {
	int age;
	struct individual *next_indiv;
};

extern double ran2(long);	/* ran2(long) is a uniform random number genrerator [0, 1) */


INDIVIDUAL *make_new_indiv(int), *add_next_indiv(INDIVIDUAL*, int);
void delete_list(INDIVIDUAL*), display_list(INDIVIDUAL*), write_to_file(FILE*, INDIVIDUAL*);
void merge_list(INDIVIDUAL*, INDIVIDUAL*), survive(INDIVIDUAL *list, double rate, long seed);
int number_of_indivs(INDIVIDUAL*);

void initialialize(), process(), finish(), setup(), age(INDIVIDUAL*);
INDIVIDUAL *reproduce_young(INDIVIDUAL*, long);
int number_offspring(long);

INDIVIDUAL *list_adult_head, *list_young_head;

long seed;
FILE *fp;
time_t time1, time2;

int main(int argc,const char *argv[])
{

	int size, trial, count = 0;

	setup();
	
	for(trial=0; trial<TRIAL; trial++){

		initialialize();
		process();
		size = number_of_indivs(list_adult_head);
		printf("Trial %d gives pop size %d\n", trial, size);
		if( size == 0 ) count++;
		delete_list(list_adult_head);
	
	} /* end of for trial */
	
	printf("Extinction %d times out of %d\n", count, TRIAL);
	finish();

	return 0;

}/* end of main */


void setup()
{
	seed = (long)time(NULL);
	ran2(-seed);	/* initialization of ran2(long) */

	time(&time1);
	fp = fopen("data_GWP","w");
}

void initialialize()
{
	int i;
	INDIVIDUAL *indiv;

	list_adult_head = make_new_indiv(0);
	indiv = list_adult_head;
	for(i = 0; i < INIT_POP_SIZE; i++)
		indiv = add_next_indiv(indiv, 0);
}

void process(long seed)
{
	int t, pop_size;
	
	for(t = 0; t < STEP; t++){
	
		/* Count the population size */
		pop_size = number_of_indivs(list_adult_head);
		fprintf(fp, "%d ", pop_size);
		/* Let all individuals reproduce */
		list_young_head = reproduce_young(list_adult_head, seed);
		/* Let all individuals survive and age one year */
		survive(list_adult_head, SURVIVAL, seed);
		age(list_adult_head);
		/* Merge existing individuals and newly recruited individuals into one list */
		merge_list(list_adult_head, list_young_head);
	
	} /* end of for t */
	
	fprintf(fp, "\n");
}

void finish()
{
	fclose(fp);

	time(&time2);
	printf("Started at %s", ctime(&time1));
	printf("Finished at %s", ctime(&time2));
	printf("Elapsed time is %ld seconds\n", time2 - time1);
}

int number_offspring(long seed)
{
	double ran;
	int num;
	
	ran = ran2(seed);

	if(ran < p0)
		num = 0;
	else if(ran < p0+p1)
		num = 1;
	else if(ran < p0+p1+p2)
		num = 2;
	else if(ran < p0+p1+p2+p3)
		num = 3;

	return num;
}

INDIVIDUAL *reproduce_young(INDIVIDUAL *list, long seed)
{
	INDIVIDUAL *young_head = make_new_indiv(0);
	INDIVIDUAL *indiv, *young;
	int num, i;
		
	young = young_head;
	for(indiv = list->next_indiv; indiv !=NULL; indiv = indiv->next_indiv){
		num = number_offspring(seed);
		for(i = 0; i < num; i++)
			young = add_next_indiv(young, 0);
	}
	
	return young_head;
}

void age(INDIVIDUAL *list)
{
	INDIVIDUAL *indiv;
	
	for(indiv = list->next_indiv; indiv != NULL; indiv = indiv->next_indiv)
		indiv->age++;
}




int number_of_indivs(INDIVIDUAL *list)
{
	INDIVIDUAL *indiv;
    int count = 0;
    
    for(indiv = list->next_indiv; indiv != NULL; indiv = indiv->next_indiv){
        count++;
    }

    return count;
}

INDIVIDUAL *make_new_indiv(int age)
{
    INDIVIDUAL *new_indiv;
    
    new_indiv = (INDIVIDUAL *)malloc(sizeof(INDIVIDUAL));
    new_indiv->age = age;
    new_indiv->next_indiv = NULL;

    return new_indiv;
}

INDIVIDUAL *add_next_indiv(INDIVIDUAL *indiv, int age)
{
    INDIVIDUAL *new_indiv;

    if(indiv->next_indiv != NULL)     return NULL;

    new_indiv = make_new_indiv(age);
    indiv->next_indiv = new_indiv;

    return new_indiv;
}

void delete_list(INDIVIDUAL *list)
{
    INDIVIDUAL *indiv, *next_indiv;

    indiv = list;
    
    do{
        next_indiv = indiv->next_indiv;
        free(indiv);
    }while( (indiv = next_indiv) != NULL );
}

void merge_list(INDIVIDUAL *list1, INDIVIDUAL *list2)
{
	INDIVIDUAL *indiv;
	
	for(indiv = list1; indiv->next_indiv != NULL; indiv = indiv->next_indiv);

	indiv->next_indiv = list2->next_indiv;
	
	free(list2);
}

void display_list(INDIVIDUAL *list)
{
    int count = 0;
    INDIVIDUAL *indiv;
    
    printf("The list is \n");
    for( indiv = list->next_indiv; indiv != NULL; indiv = indiv->next_indiv){
        count++;
        printf("%d\n", indiv->age);
    }
    printf("The number of individuals is %d\n\n", count);
}

void write_to_file(FILE *fp, INDIVIDUAL *list)
{
	INDIVIDUAL *indiv;
	
    for(indiv = list->next_indiv; indiv != NULL; indiv = indiv->next_indiv)
        fprintf(fp, "%d ", indiv->age);
    fprintf(fp,"\n");
}

void survive(INDIVIDUAL *list,double rate,long seed)
{
	INDIVIDUAL *prev, *target, *next;
	
	prev=list;
	target=prev->next_indiv;
	
	while(target !=NULL){
	
		if(ran2(seed) > rate ){
		
		prev -> next_indiv = target -> next_indiv;
		next=target->next_indiv;
		free(target);
		target =next;
		
		}
		else{
		target = target -> next_indiv;
		prev = prev->next_indiv;
		}
	}
}
